"""
This module contains error classes for translating various gRPC server response codes into more pythonic exceptions
"""

from collections.abc import Callable, Coroutine
from typing import Any, Protocol, TypeVar, cast

from grpc import RpcError, StatusCode
from grpc.aio import AioRpcError


class AuthenticationError(IOError):
    """AuthenticationError indicates that a server request failed due to a missing or invalid authentication token"""


class ArgumentError(ValueError):
    """ArgumentError indicates that a server request failed due to missing or invalid arguments"""


class NotFoundError(KeyError):
    """NotFoundError indicates that a given resource was not found on the server side"""


class SubscriptionLimitExceededError(IOError):
    """SubscriptionLimitExceededError indicates that the subscription limit for a resource has been reached"""


class InternalServerError(KeyError):
    """InternalServerError indicates that an unexpected error happened on the server side"""


Stub = TypeVar("Stub")


def with_pythonic_errors(stub: Stub, async_funcs: bool = False) -> Stub:
    """
    Wrap a sync gRPC stub to translate rpc errors into pythonic exceptions.

    This is done this way instead of an error handling interceptor, because tracebacks are much easier to read when
    they are raised directly at the location of the RPC method call.

    Args:
        stub: The grpc stub to wrap.
        async_funcs: Whether to wrap the callables as coroutines or not. Defaults to False.

    Returns:
        The stub with the rpc methods wrapped.
    """
    wrap_func = _wrap_rpc if not async_funcs else _async_wrap_rpc
    for name, rpc in stub.__dict__.items():
        if callable(rpc):
            setattr(stub, name, wrap_func(rpc))  # type: ignore[assignment]
    return stub


class AnyRpcError(Protocol):
    """Protocol for gRPC errors that works for both sync and async gRPC."""

    def code(self) -> StatusCode: ...
    def details(self) -> str: ...


def translate_rpc_error(err: AnyRpcError) -> Exception:
    # translate specific error codes to more pythonic errors
    match err.code():
        case StatusCode.UNAUTHENTICATED:
            return AuthenticationError(f"Unauthenticated: {err.details()}")
        case StatusCode.PERMISSION_DENIED:
            return AuthenticationError(f"Unauthorized: {err.details()}")
        case StatusCode.NOT_FOUND:
            return NotFoundError(err.details())
        case StatusCode.RESOURCE_EXHAUSTED:
            return SubscriptionLimitExceededError(err.details())
        case StatusCode.INVALID_ARGUMENT:
            return ArgumentError(err.details())

    return InternalServerError(f"Oops, something went wrong: {err.details()}")


def _wrap_rpc(rpc: Callable[[Any], Any]) -> Callable[[Any], Any]:
    def call(*args: Any, **kwargs: Any) -> Any:
        try:
            return rpc(*args, **kwargs)
        except (RpcError, AioRpcError) as err:
            error = translate_rpc_error(cast(AnyRpcError, err))

            # raise the appropriate exception for the error code we received
            raise error from None

    return call


def _async_wrap_rpc(rpc: Callable[[Any], Coroutine[Any, Any, Any]]) -> Callable[[Any], Coroutine[Any, Any, Any]]:
    async def call(*args: Any, **kwargs: Any) -> Any:
        try:
            return await rpc(*args, **kwargs)
        except (RpcError, AioRpcError) as err:
            error = translate_rpc_error(cast(AnyRpcError, err))

            # raise the appropriate exception for the error code we received
            raise error from None

    return call



from collections.abc import AsyncIterator, Awaitable, Callable

from _tilebox.grpc.pagination import AnyResultPage, Pagination


async def paginated_request(
    paging_request: Callable[[Pagination], Awaitable[AnyResultPage]],
    initial_page: Pagination,
) -> AsyncIterator[AnyResultPage]:
    """Make a paginated request to a gRPC service endpoint.

    The endpoint is expected to return a next_page field, which is used for subsequent requests. Once no such
    next_page field is returned, the request is completed.

    Args:
        paging_request: A function that takes a page as input and returns a Datapoints object
            Often this will be a functools.partial object that wraps a gRPC service endpoint
            and only leaves the page argument remaining
        initial_page: The initial page to request

    Yields:
        Datapoints: The individual pages of the response
    """
    response = await paging_request(initial_page)
    yield response

    while response.next_page.starting_after is not None:
        response = await paging_request(response.next_page)
        yield response



from collections.abc import Callable, Iterator
from typing import Protocol, TypeVar
from uuid import UUID


class Pagination(Protocol):
    limit: int | None
    starting_after: UUID | None


class ResultPage(Protocol):
    @property
    def next_page(self) -> Pagination: ...


AnyResultPage = TypeVar("AnyResultPage", bound=ResultPage)


def paginated_request(
    paging_request: Callable[[Pagination], AnyResultPage],
    initial_page: Pagination,
) -> Iterator[AnyResultPage]:
    """Make a paginated request to a gRPC service endpoint.

    The endpoint is expected to return a next_page field, which is used for subsequent requests. Once no such
    next_page field is returned, the request is completed.

    Args:
        paging_request: A function that takes a page as input and returns a Datapoints object
            Often this will be a functools.partial object that wraps a gRPC service endpoint
            and only leaves the page argument remaining
        initial_page: The initial page to request

    Yields:
        Datapoints: The individual pages of the response
    """
    response = paging_request(initial_page)
    yield response

    while response.next_page.starting_after is not None:
        response = paging_request(response.next_page)
        yield response



from collections.abc import Callable
from typing import Any, TypeVar, cast

from _tilebox.grpc.error import AnyRpcError, translate_rpc_error
from grpc import RpcError
from grpc.aio import AioRpcError

Stub = TypeVar("Stub")


def with_pythonic_errors(stub: Stub) -> Stub:
    """
    Wrap an async gRPC stub to translate rpc errors into pythonic exceptions.

    This is done this way instead of an error handling interceptor, because tracebacks are much easier to read when
    they are raised directly at the location of the RPC method call.

    Args:
        stub: The async grpc stub to wrap.

    Returns:
        The stub with the rpc methods wrapped.
    """
    for name, rpc in stub.__dict__.items():
        if callable(rpc):
            setattr(stub, name, _wrap_rpc(rpc))
    return stub


def _wrap_rpc(rpc: Callable[[Any], Any]) -> Callable[[Any], Any]:
    async def call(*args: Any, **kwargs: Any) -> Any:
        try:
            return await rpc(*args, **kwargs)
        except (RpcError, AioRpcError) as err:
            error = translate_rpc_error(cast(AnyRpcError, err))

            # raise the appropriate exception for the error code we received
            raise error from None

    return call



from collections.abc import Callable, Sequence
from typing import TypeVar

from _tilebox.grpc.channel import CHANNEL_OPTIONS, ChannelInfo, ChannelProtocol, add_metadata, parse_channel_info
from grpc import Compression, ssl_channel_credentials
from grpc.aio import (
    Channel,
    ClientCallDetails,
    ClientInterceptor,
    UnaryUnaryCall,
    UnaryUnaryClientInterceptor,
    insecure_channel,
    secure_channel,
)


def open_channel(url: str, auth_token: str | None = None) -> Channel:
    """Open an async gRPC channel to the given URL.

    Args:
        url: The URL to open a channel to. Depending on the URL, the channel will be a secure (SSL) or insecure channel.
        auth_token: Authentication token for the channel. If set, an interceptor channel will be created which adds
            the given token as metadata to each request.

    Returns:
        A gRPC channel.
    """
    channel_info = parse_channel_info(url)
    interceptors: list[ClientInterceptor] = []
    if auth_token is not None:
        interceptors = [_AuthMetadataInterceptor(auth_token), *interceptors]  # add auth interceptor as the first one

    return _open_channel(channel_info, interceptors)


def _open_channel(channel_info: ChannelInfo, interceptors: Sequence[ClientInterceptor]) -> Channel:
    match channel_info.protocol:
        case ChannelProtocol.HTTPS:
            return secure_channel(
                f"{channel_info.address}:{channel_info.port}",
                ssl_channel_credentials(),
                CHANNEL_OPTIONS,
                compression=Compression.Gzip,
                interceptors=interceptors,
            )
        case ChannelProtocol.HTTP:
            return insecure_channel(
                f"{channel_info.address}:{channel_info.port}",
                CHANNEL_OPTIONS,
                compression=Compression.NoCompression,
                interceptors=interceptors,
            )
        case ChannelProtocol.UNIX:
            return insecure_channel(
                channel_info.address, CHANNEL_OPTIONS, compression=Compression.NoCompression, interceptors=interceptors
            )
        case _:
            raise ValueError(f"Unsupported channel protocol: {channel_info.protocol}")


RequestType = TypeVar("RequestType")


class _AuthMetadataInterceptor(UnaryUnaryClientInterceptor):
    def __init__(self, auth_token: str) -> None:
        """A gRPC channel interceptor which adds the authorization token as metadata to every request.

        Args:
            auth_token: The authorization token.
        """
        super().__init__()
        self._auth = ("authorization", f"Bearer {auth_token}")

    async def intercept_unary_unary(
        self,
        continuation: Callable[[ClientCallDetails, RequestType], UnaryUnaryCall],
        client_call_details: ClientCallDetails,
        request: RequestType,
    ) -> UnaryUnaryCall:
        return await continuation(add_metadata(client_call_details, [self._auth]), request)



from collections.abc import Callable, Iterator
from queue import Queue
from threading import Thread
from typing import Any, TypeVar

T = TypeVar("T")


def concurrent_producer_consumer(
    producer: Iterator[T],
    consumer: Callable[[T], None] | Callable[[T], Any],
    buffer_size: int = 10,
) -> None:
    """
    Consume messages from a producer (e.g. a service endpoint) as they come in. This ensures that the consumer
    is already computing each produced message while the next message is being produced.

    If the consumer is actually slower than the producer producing new message, the messages are buffered up to the
    given buffer size. If the buffer is full, the producer will block until the consumer has consumed a message.

    Args:
        producer: The generator to consume from
        consumer: The consumer function to call for each message
        buffer_size: The maximum number of messages to buffer before blocking the producer
    """

    queue = Queue(maxsize=buffer_size)

    # offload consuming to a separate thread, so that we can continue producing data on the main thread
    consumer_thread = Thread(target=_consumer, args=(consumer, queue))
    consumer_thread.start()
    # do this on the main thread - the producer may have callbacks (e.g. progress bar updates) and those should
    # happen on the main thread - otherwise issues may arise with libraries like streamlit
    _producer(producer, queue)
    consumer_thread.join()


def _producer(producer: Iterator[T], queue: Queue[T | None]) -> None:
    """
    Forward all messages from a producer (async iterator) and send them to a buffered send stream.
    """
    for message in producer:
        queue.put(message, block=True)

    queue.put(None, block=True)  # indicate that we are done


def _consumer(consume: Callable[[T], None] | Callable[[T], Any], queue: Queue[T | None]) -> None:
    """
    Receive messages from a buffered receive stream and forward them to a consumer until no more messages are available.
    """
    while True:
        message = queue.get(block=True)
        if message is None:
            break
        consume(message)
